from typing import Any


class Variable:
    """Mock airflow.models.Variable class, make migrator do less thing."""

    @classmethod
    def get(cls, key: str) -> str:
        """Modck airflow.models.Variable.get method, only return income parameter."""
        return key

    @classmethod
    def set(cls, key: str, value: Any) -> str:
        """Mock airflow.models.Variable.set method, do nothing."""

    @classmethod
    def update(cls, key: str, value: Any) -> str:
        """Mock airflow.models.Variable.update method, do nothing."""

    @classmethod
    def delete(cls, key: str) -> str:
        """Mock airflow.models.Variable.delete method, do nothing."""



from enum import Enum


class TriggerRule(str, Enum):
    """Mock trigger rule"""

    ALL_SUCCESS = "all_success"
    ALL_FAILED = "all_failed"
    ALL_DONE = "all_done"
    ONE_SUCCESS = "one_success"
    ONE_FAILED = "one_failed"
    ONE_DONE = "one_done"
    NONE_FAILED = "none_failed"
    NONE_FAILED_OR_SKIPPED = "none_failed_or_skipped"
    NONE_SKIPPED = "none_skipped"
    DUMMY = "dummy"
    ALWAYS = "always"
    NONE_FAILED_MIN_ONE_SUCCESS = "none_failed_min_one_success"
    ALL_SKIPPED = "all_skipped"

    @classmethod
    def is_valid(cls, trigger_rule: str) -> bool:
        """Validates a trigger rule."""
        return trigger_rule in cls.all_triggers()

    @classmethod
    def all_triggers(cls) -> set[str]:
        """Returns all trigger rules."""
        return set(cls.__members__.values())

    def __str__(self) -> str:
        return self.value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import warnings
from typing import List, Set, Union

import libcst as cst
import libcst.matchers as m
from libcst import BaseExpression, FlattenSentinel, RemovalSentinel, SimpleStatementLine
from libcst.metadata import PositionProvider, QualifiedName, QualifiedNameProvider

from air2phin.constants import KEYWORD
from air2phin.core.rules.config import Config
from air2phin.core.transformer.imports import ImportTransformer
from air2phin.core.transformer.operators import OpTransformer


class Transformer(cst.CSTTransformer):
    """CST Transformer route class from airflow to dolphinscheduler-sdk-python.

    The main class to call each rules to migrate, just like a router, currently will route to `imports` and
    `operators` transformer.

    :param config: libCST transformer configuration, use it to get importer and callable migrate setting.
    """

    METADATA_DEPENDENCIES = (
        QualifiedNameProvider,
        PositionProvider,
    )

    def __init__(self, config: Config):
        super().__init__()
        self.config: Config = config
        self.workflow_alias = set()
        self.have_submit_expr = set()

    @staticmethod
    def _get_qualified_name(qualifie: Set[QualifiedName]) -> str:
        if len(qualifie) > 1:
            warnings.warn(
                "QualifiedNameProvider get more than one qualified name, will use the first one.",
                RuntimeWarning,
            )
        for q in qualifie:
            return q.name

    # @m.call_if_inside(
    #     m.Call(
    #         func=m.MatchMetadataIfTrue(
    #             meta.QualifiedNameProvider,
    #             lambda qualnames: any(
    #                 n.name in call_cov for n in qualnames
    #             ),
    #         )
    #     )
    # )
    def leave_Call(
        self, original_node: cst.Call, updated_node: cst.Call
    ) -> BaseExpression:
        qnp = self.get_metadata(QualifiedNameProvider, original_node)
        if qnp:
            qnp_name = self._get_qualified_name(qnp)
            if qnp_name in self.config.calls:
                return updated_node.visit(OpTransformer(self.config, qnp_name))
        return updated_node

    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> Union[
        cst.BaseSmallStatement, FlattenSentinel[cst.BaseSmallStatement], RemovalSentinel
    ]:
        """Migrate from import statement."""
        return updated_node.visit(ImportTransformer(self.config))

    def leave_WithItem_asname(self, node: cst.WithItem) -> None:
        """Get airflow Dags alias names."""
        if m.matches(node.item, m.Call()) and m.matches(
            cst.ensure_type(node.item, cst.Call).func,
            m.Name(value=KEYWORD.AIRFLOW_DAG_SIMPLE),
        ):
            self.workflow_alias.add(node.asname.name.value)

    def leave_Expr(
        self, original_node: cst.Expr, updated_node: cst.Expr
    ) -> Union[
        cst.BaseSmallStatement,
        cst.FlattenSentinel[cst.BaseSmallStatement],
        RemovalSentinel,
    ]:
        """Update workflow ``alias.submit()`` expr exits or not statement."""
        if m.matches(
            original_node.value,
            m.Call(
                func=m.Attribute(
                    value=m.OneOf(*[m.Name(a) for a in self.workflow_alias]),
                    attr=m.Name(KEYWORD.WORKFLOW_SUBMIT),
                )
            ),
        ):
            self.have_submit_expr.add(original_node.value.func.value.value)
        return updated_node

    def _build_submit_exprs(self) -> List[SimpleStatementLine]:
        miss_alias = self.workflow_alias.difference(self.have_submit_expr)
        return [
            cst.parse_statement(f"{alias}.{KEYWORD.WORKFLOW_SUBMIT}()")
            for alias in miss_alias
        ]

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        if self.have_submit_expr == self.workflow_alias:
            return updated_node

        # add submit expr when do not have
        body_with_submit = list(updated_node.body)
        body_with_submit.extend(self._build_submit_exprs())
        return updated_node.with_changes(body=body_with_submit)



from air2phin.fake.core.connection import Connection
from air2phin.fake.core.hook import BaseHook


class PostgresHook(BaseHook):
    """Interact with PostgresSQL.

    This hook is a fake hook for Airflow Postgres hook, to allow user do not change their code but use
    dolphinscheduler datasource connection to connect to PostgresSQL database.

    :param connection: specific hook connection. :class:``air2phin.fake.core.connection.Connection`` object.
    """

    def __init__(self, connection: Connection, *args, **kwargs):
        super().__init__(connection)

    def get_conn(self) -> "psycopg2.extensions.connection":  # noqa: F821
        """Get postgres connection object."""
        try:
            import psycopg2
        except ImportError:
            raise ImportError("This psycopg2 module does not seem to be installed.")

        conn_args = dict(
            host=self.connection.host,
            port=self.connection.port,
            dbname=self.connection.schema,
            user=self.connection.login,
            password=self.connection.password,
        )
        return psycopg2.connect(**conn_args)



import json
import os
import re
from contextlib import closing
from typing import Any, Callable, Iterable, Mapping, Tuple, Union

import sqlparse
from sqlalchemy import create_engine, text

from air2phin.constants import TOKEN
from air2phin.fake.core.connection import Connection


def fetch_all_handler(cursor) -> list[tuple] | None:
    """Handler for DbApiHook.run() to return results."""
    if cursor.description is not None:
        return cursor.fetchall()
    else:
        return None


class BaseHook:
    """Base hook for all fake hook.

    This hook is a fake hook for Airflow base hook, to allow user do not change their code but use
    dolphinscheduler datasource connection to connect specific datasource.

    :param connection: specific hook connection. :class:``air2phin.fake.core.connection.Connection`` object.
    """

    def __init__(self, connection: Connection):
        self.connection = connection

    def get_conn(self):
        """Get hook connection object, depend on subclass return type."""
        raise NotImplementedError

    @staticmethod
    def parser_conn_namedtuple(connection_params: str) -> Connection:
        """Parse dolphinscheduler connection_params to Connection.

        :param connection_params: connection_params from dolphinscheduler datasource.
        """
        data = json.loads(connection_params)

        pattern = re.compile(
            "jdbc:.*://(?P<host>[\\w\\W]+):(?P<port>\\d+)/(?P<database>\\w+)(\\?|$)"
        )
        # node name change to url when using seatunnel datasource connector
        try:
            pattern_match = pattern.match(
                data.get("jdbcUrl", data.get("url", None))
            ).groupdict()
        except Exception:
            raise ValueError(f"Can not parser connection params: {connection_params}.")

        return Connection(
            host=pattern_match.get("host", None),
            port=int(pattern_match.get("port", None)),
            schema=pattern_match.get("database", None),
            login=data.get("user", None),
            password=data.get("password", None),
        )

    @staticmethod
    def _get_type_name(conn_id) -> Tuple[Any, str]:
        if TOKEN.POINT in conn_id:
            return conn_id.strip().split(TOKEN.POINT)
        return None, conn_id.strip()

    @classmethod
    def _get_connection_params_from_env(
        cls, metadata_conn: str, conn_id: str
    ) -> Connection:
        sql_qry_type_name = (
            "SELECT connection_params FROM t_ds_datasource WHERE type = {type} and "
            "name = '{name}'"
        )
        sql_qry_name = (
            "SELECT connection_params FROM t_ds_datasource WHERE name = '{name}'"
        )

        database_type_map = dict(
            mysql=0,
            postgresql=1,
            hive=2,
            spark=3,
            clickhouse=4,
            oracle=5,
            sqlserver=6,
            db2=7,
            presto=8,
            h2=9,
            redshift=10,
            dameng=11,
            starrocks=12,
        )

        datasource_type, datasource_name = cls._get_type_name(conn_id)
        engine = create_engine(metadata_conn, echo=True)

        with engine.connect() as conn:
            # conn_id not in format of datasource_type.datasource_name
            if TOKEN.POINT not in conn_id:
                result_name = conn.execute(
                    text(sql_qry_name.format(name=datasource_name))
                )
                if result_name.rowcount == 0:
                    raise ValueError(
                        f"Connection {conn_id} not found in dolphinscheduler metadata database."
                    )
                elif result_name.rowcount > 1:
                    raise ValueError(
                        f"Connection {conn_id} is not unique in dolphinscheduler metadata database, please "
                        f"use ``datasource_type.datasource_name`` to specify."
                    )
                record = result_name.fetchone()
                return cls.parser_conn_namedtuple(record[0])

            # conn_id in format of datasource_type.datasource_name
            if datasource_type.lower() not in database_type_map:
                raise ValueError(
                    f"Datasource type `{datasource_type}` not support currently, please use one of "
                    f"{list(database_type_map.keys())}"
                )
            result_type_name = conn.execute(
                text(
                    sql_qry_type_name.format(
                        type=database_type_map.get(datasource_type),
                        name=datasource_name,
                    )
                )
            )
            if result_type_name.rowcount == 0:
                raise ValueError(
                    f"Connection {conn_id} not found in dolphinscheduler metadata database."
                )
            record = result_type_name.fetchone()
            return cls.parser_conn_namedtuple(record[0])

    @classmethod
    def get_connection(cls, conn_id: str) -> Connection:
        """Get connection from dolphinscheduler metadata database.

        This method is a fake function for Airflow connection get_connection, to allow user do not change
        their code but use dolphinscheduler datasource and return
        :class:``air2phin.fake.core.connection.Connection`` object.

        :param conn_id: connection id, if in format of datasource_type.datasource_name, will query by type
            and name, and if only use datasource_name, will query by name only.
        """
        try:
            from pydolphinscheduler.models.datasource import Datasource

            datasource_type, datasource_name = cls._get_type_name(conn_id)
            database: Datasource = Datasource.get(
                datasource_name=datasource_name, datasource_type=datasource_type
            )
            return cls.parser_conn_namedtuple(database.connection_params)
        except ImportError:
            metadata_conn = os.environ.get("AIR2PHIN_FAKE_CONNECTION", None)
            if metadata_conn is None:
                raise ValueError(
                    "Can not get dolphinscheduler metadata connection information, neither package"
                    "``pydolphinscheduler`` installed nor environment variable ``AIR2PHIN_FAKE_CONNECTION``"
                    "is set, please do one of them to keep going."
                )
            return cls._get_connection_params_from_env(metadata_conn, conn_id)

    def _run_command(self, cur, sql_statement, parameters):
        """Execute command."""
        if parameters:
            cur.execute(sql_statement, parameters)
        else:
            cur.execute(sql_statement)

    def run(
        self,
        sql: Union[str, Iterable[str]],
        autocommit: bool = False,
        parameters: Union[Iterable, Mapping, None] = None,
        handler: Union[Callable, None] = None,
        split_statements: bool = False,
        return_last: bool = True,
    ) -> Union[Any, list[Any], None]:
        """Mock sql run command."""
        if isinstance(sql, str):
            if split_statements:
                splits = sqlparse.split(sqlparse.format(sql, strip_comments=True))
                sql: list[str] = list(filter(None, splits))
            else:
                sql = [sql]

        if not sql:
            raise ValueError("List of SQL statements is empty")

        with closing(self.get_conn()) as conn:
            with closing(conn.cursor()) as cur:
                results = []
                for sql_statement in sql:
                    self._run_command(cur, sql_statement, parameters)

                    if handler is not None:
                        result = handler(cur)
                        results.append(result)

            if autocommit:
                conn.commit()

        if handler is None:
            return None
        elif isinstance(sql, str) and return_last:
            return results[-1]
        else:
            return results

    @staticmethod
    def fetch_all_handler(cursor) -> Union[list[tuple], None]:
        """Handler for DbApiHook.run() to return results."""
        if cursor.description is not None:
            return cursor.fetchall()
        else:
            return None

    def get_records(
        self,
        sql: Union[str, list[str]],
        parameters: Union[Iterable, Mapping, None] = None,
    ) -> Any:
        """Mock executes sql and returns records."""
        return self.run(sql=sql, parameters=parameters, handler=fetch_all_handler)
